package tree

import (
	"context"
	"crypto/ecdsa"
	"crypto/sha256"
	"encoding/hex"
	"errors"
	"fmt"
	"testing"

	aclV2 "git.frostfs.info/TrueCloudLab/frostfs-api-go/v2/acl"
	"git.frostfs.info/TrueCloudLab/frostfs-contract/frostfsid/client"
	containercore "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/core/container"
	frostfsidcore "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/core/frostfsid"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/core/netmap"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/util/logger/test"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/ape"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/bearer"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/container"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/container/acl"
	cid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/container/id"
	cidtest "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/container/id/test"
	netmapSDK "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/netmap"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/user"
	"git.frostfs.info/TrueCloudLab/policy-engine/pkg/chain"
	policyengine "git.frostfs.info/TrueCloudLab/policy-engine/pkg/engine"
	"git.frostfs.info/TrueCloudLab/policy-engine/pkg/engine/inmemory"
	nativeschema "git.frostfs.info/TrueCloudLab/policy-engine/schema/native"
	"github.com/nspcc-dev/neo-go/pkg/crypto/keys"
	"github.com/nspcc-dev/neo-go/pkg/util"
	"github.com/stretchr/testify/require"
)

type dummyNetmapSource struct {
	netmap.Source
}

type dummyContainerSource map[string]*containercore.Container

func (s dummyContainerSource) List() ([]cid.ID, error) {
	res := make([]cid.ID, 0, len(s))
	var cnr cid.ID

	for cidStr := range s {
		err := cnr.DecodeString(cidStr)
		if err != nil {
			return nil, err
		}

		res = append(res, cnr)
	}

	return res, nil
}

func (s dummyContainerSource) Get(id cid.ID) (*containercore.Container, error) {
	cnt, ok := s[id.String()]
	if !ok {
		return nil, errors.New("container not found")
	}
	return cnt, nil
}

func (s dummyContainerSource) DeletionInfo(id cid.ID) (*containercore.DelInfo, error) {
	return &containercore.DelInfo{}, nil
}

type dummyEACLSource map[string]*containercore.EACL

func (s dummyEACLSource) GetEACL(id cid.ID) (*containercore.EACL, error) {
	cntEACL, ok := s[id.String()]
	if !ok {
		return nil, errors.New("container not found")
	}
	return cntEACL, nil
}

func testContainer(owner user.ID) container.Container {
	var r netmapSDK.ReplicaDescriptor
	r.SetNumberOfObjects(1)

	var pp netmapSDK.PlacementPolicy
	pp.AddReplicas(r)

	var cnt container.Container
	cnt.SetOwner(owner)
	cnt.SetPlacementPolicy(pp)

	return cnt
}

type dummyFrostfsIDSubjProvider struct {
}

var _ frostfsidcore.SubjectProvider = (*dummyFrostfsIDSubjProvider)(nil)

func (p *dummyFrostfsIDSubjProvider) GetSubject(addr util.Uint160) (*client.Subject, error) {
	return &client.Subject{}, nil
}

func (p *dummyFrostfsIDSubjProvider) GetSubjectExtended(addr util.Uint160) (*client.SubjectExtended, error) {
	return &client.SubjectExtended{}, nil
}

type dummyState struct{}

func (d *dummyState) CurrentEpoch() uint64 {
	return 0
}

func TestMessageSign(t *testing.T) {
	privs := make([]*keys.PrivateKey, 4)
	for i := range privs {
		p, err := keys.NewPrivateKey()
		require.NoError(t, err)
		privs[i] = p
	}

	cid1 := cidtest.ID()
	cid2 := cidtest.ID()

	var ownerID user.ID
	user.IDFromKey(&ownerID, (ecdsa.PublicKey)(*privs[0].PublicKey()))

	cnr := &containercore.Container{
		Value: testContainer(ownerID),
	}

	localStorageChainMock := inmemory.NewInmemoryLocalStorage()
	morphStorageChainMock := inmemory.NewInmemoryMorphRuleChainStorage()

	s := &Service{
		cfg: cfg{
			log:      test.NewLogger(t),
			key:      &privs[0].PrivateKey,
			nmSource: dummyNetmapSource{},
			cnrSource: dummyContainerSource{
				cid1.String(): cnr,
			},
			state:                    &dummyState{},
			eaclSource:               dummyEACLSource{},
			router:                   policyengine.NewDefaultChainRouterWithLocalOverrides(morphStorageChainMock, localStorageChainMock),
			frostfsidSubjectProvider: &dummyFrostfsIDSubjProvider{},
		},
	}

	rawCID1 := make([]byte, sha256.Size)
	cid1.Encode(rawCID1)

	req := &MoveRequest{
		Body: &MoveRequest_Body{
			ContainerId: rawCID1,
			ParentId:    1,
			NodeId:      2,
			Meta: []*KeyValue{
				{Key: "kkk", Value: []byte("vvv")},
			},
		},
	}

	const treeID = "version"

	t.Run("missing signature, no panic", func(t *testing.T) {
		require.Error(t, s.verifyClient(context.Background(), req, cid2, nil, nativeschema.MethodMoveTreeNode, treeID))
	})

	require.NoError(t, SignMessage(req, &privs[0].PrivateKey))

	t.Run("deprecated basic ACL scheme", func(t *testing.T) {
		t.Run("zero-filled bACL", func(t *testing.T) {
			require.Error(t, s.verifyClient(context.Background(), req, cid1, nil, nativeschema.MethodMoveTreeNode, treeID))
		})

		t.Run("non-zero-filled bACL", func(t *testing.T) {
			cnr.Value.SetBasicACL(acl.PublicRW)
			require.Error(t, s.verifyClient(context.Background(), req, cid1, nil, nativeschema.MethodMoveTreeNode, treeID))
		})
	})

	localStorageChainMock.AddOverride(chain.Ingress, policyengine.ContainerTarget(cid1.EncodeToString()), &chain.Chain{
		Rules: []chain.Rule{
			{
				Status: chain.Allow,
				Actions: chain.Actions{
					Names: []string{nativeschema.MethodMoveTreeNode},
				},
				Resources: chain.Resources{
					Names: []string{
						nativeschema.ResourceFormatRootTrees,
					},
				},
				Condition: []chain.Condition{
					{
						Op:    chain.CondStringEquals,
						Kind:  chain.KindRequest,
						Key:   nativeschema.PropertyKeyActorRole,
						Value: nativeschema.PropertyValueContainerRoleOwner,
					},
					{
						Op:    chain.CondStringEquals,
						Kind:  chain.KindRequest,
						Key:   nativeschema.PropertyKeyActorPublicKey,
						Value: hex.EncodeToString(privs[0].PublicKey().Bytes()),
					},
				},
			},
			{
				Status: chain.Allow,
				Actions: chain.Actions{
					Names: []string{nativeschema.MethodGetSubtree},
				},
				Resources: chain.Resources{
					Names: []string{
						nativeschema.ResourceFormatRootTrees,
					},
				},
				Condition: []chain.Condition{},
			},
		},
	})

	require.NoError(t, s.verifyClient(context.Background(), req, cid1, nil, nativeschema.MethodMoveTreeNode, treeID))

	t.Run("invalid CID", func(t *testing.T) {
		require.Error(t, s.verifyClient(context.Background(), req, cid2, nil, nativeschema.MethodMoveTreeNode, treeID))
	})

	t.Run("invalid key", func(t *testing.T) {
		require.NoError(t, SignMessage(req, &privs[1].PrivateKey))
		require.Error(t, s.verifyClient(context.Background(), req, cid1, nil, nativeschema.MethodMoveTreeNode, treeID))
	})

	t.Run("bearer", func(t *testing.T) {
		t.Run("invalid bearer", func(t *testing.T) {
			req.Body.BearerToken = []byte{0xDE, 0xAD, 0xBE, 0xEF}
			require.NoError(t, SignMessage(req, &privs[0].PrivateKey))
			require.ErrorContains(t, s.verifyClient(context.Background(), req, cid1, req.GetBody().GetBearerToken(), nativeschema.MethodMoveTreeNode, treeID), "cannot parse invalid wire-format data")
		})

		t.Run("invalid bearer CID", func(t *testing.T) {
			bt := testBearerToken(cid2, privs[1].PublicKey(), privs[2].PublicKey()) //testBearerToken(cid2, privs[1].PublicKey(), privs[2].PublicKey())
			require.NoError(t, bt.Sign(privs[0].PrivateKey))
			req.Body.BearerToken = bt.Marshal()

			require.NoError(t, SignMessage(req, &privs[1].PrivateKey))
			require.ErrorContains(t, s.verifyClient(context.Background(), req, cid1, req.GetBody().GetBearerToken(), nativeschema.MethodMoveTreeNode, treeID), errBearerInvalidContainerID.Error())
		})

		t.Run("invalid bearer owner", func(t *testing.T) {
			bt := testBearerToken(cid1, privs[1].PublicKey(), privs[2].PublicKey()) //testBearerToken(cid1, privs[1].PublicKey(), privs[2].PublicKey())
			require.NoError(t, bt.Sign(privs[1].PrivateKey))
			req.Body.BearerToken = bt.Marshal()

			require.NoError(t, SignMessage(req, &privs[1].PrivateKey))
			require.ErrorContains(t, s.verifyClient(context.Background(), req, cid1, req.GetBody().GetBearerToken(), nativeschema.MethodMoveTreeNode, treeID), errBearerNotSignedByOwner.Error())
		})

		t.Run("invalid bearer signature", func(t *testing.T) {
			bt := testBearerToken(cid1, privs[1].PublicKey(), privs[2].PublicKey())
			require.NoError(t, bt.Sign(privs[0].PrivateKey))

			var bv2 aclV2.BearerToken
			bt.WriteToV2(&bv2)
			bv2.GetSignature().SetSign([]byte{1, 2, 3})
			req.Body.BearerToken = bv2.StableMarshal(nil)

			require.NoError(t, SignMessage(req, &privs[1].PrivateKey))
			require.ErrorContains(t, s.verifyClient(context.Background(), req, cid1, req.GetBody().GetBearerToken(), nativeschema.MethodMoveTreeNode, treeID), errBearerInvalidSignature.Error())
		})

		t.Run("impersonate", func(t *testing.T) {
			var bt bearer.Token

			bt.SetImpersonate(true)
			require.NoError(t, bt.Sign(privs[1].PrivateKey))

			req.Body.BearerToken = bt.Marshal()
			require.NoError(t, SignMessage(req, &privs[0].PrivateKey))

			require.ErrorContains(t, s.verifyClient(context.Background(), req, cid1, req.GetBody().GetBearerToken(), nativeschema.MethodMoveTreeNode, treeID), "access to object operation denied")
			require.NoError(t, s.verifyClient(context.Background(), req, cid1, req.GetBody().GetBearerToken(), nativeschema.MethodGetSubtree, treeID))
		})

		bt := testBearerToken(cid1, privs[1].PublicKey(), privs[2].PublicKey())
		require.NoError(t, bt.Sign(privs[0].PrivateKey))
		req.Body.BearerToken = bt.Marshal()
		cnr.Value.SetBasicACL(acl.PublicRWExtended)

		t.Run("put and get", func(t *testing.T) {
			require.NoError(t, SignMessage(req, &privs[1].PrivateKey))
			require.NoError(t, s.verifyClient(context.Background(), req, cid1, req.GetBody().GetBearerToken(), nativeschema.MethodMoveTreeNode, treeID))
			require.NoError(t, s.verifyClient(context.Background(), req, cid1, req.GetBody().GetBearerToken(), nativeschema.MethodGetSubtree, treeID))
		})
		t.Run("only get", func(t *testing.T) {
			require.NoError(t, SignMessage(req, &privs[2].PrivateKey))
			require.Error(t, s.verifyClient(context.Background(), req, cid1, req.GetBody().GetBearerToken(), nativeschema.MethodMoveTreeNode, treeID))
			require.NoError(t, s.verifyClient(context.Background(), req, cid1, req.GetBody().GetBearerToken(), nativeschema.MethodGetSubtree, treeID))
		})
		t.Run("none", func(t *testing.T) {
			require.NoError(t, SignMessage(req, &privs[3].PrivateKey))
			require.Error(t, s.verifyClient(context.Background(), req, cid1, req.GetBody().GetBearerToken(), nativeschema.MethodMoveTreeNode, treeID))
			require.Error(t, s.verifyClient(context.Background(), req, cid1, req.GetBody().GetBearerToken(), nativeschema.MethodGetSubtree, treeID))
		})
	})
}

func testBearerToken(cid cid.ID, forPutGet, forGet *keys.PublicKey) bearer.Token {
	var b bearer.Token
	b.SetExp(100)

	allowsGetOpForAnyRoleForBothKeys := chain.Chain{
		Rules: []chain.Rule{
			{
				Status: chain.Allow,
				Actions: chain.Actions{
					Names: []string{nativeschema.MethodGetTreeNodeByPath,
						nativeschema.MethodGetSubtree,
						nativeschema.MethodListTrees},
				},
				Resources: chain.Resources{
					Names: []string{
						nativeschema.ResourceFormatRootTrees,
					},
				},
				Condition: []chain.Condition{
					{
						Op:    chain.CondSliceContains,
						Kind:  chain.KindRequest,
						Key:   nativeschema.PropertyKeyActorPublicKey,
						Value: fmt.Sprintf("%s\x00%s", hex.EncodeToString(forPutGet.Bytes()), hex.EncodeToString(forGet.Bytes())),
					},
				},
			},
		},
	}

	allowsPutOpForAnyRoleForOneKeyOnly := chain.Chain{
		Rules: []chain.Rule{
			{
				Status: chain.Allow,
				Actions: chain.Actions{
					Names: []string{nativeschema.MethodAddTreeNode,
						nativeschema.MethodAddTreeNodeByPath,
						nativeschema.MethodRemoveTreeNode,
						nativeschema.MethodMoveTreeNode,
					},
				},
				Resources: chain.Resources{
					Names: []string{
						nativeschema.ResourceFormatRootTrees,
					},
				},
				Condition: []chain.Condition{
					{
						Op:    chain.CondStringEquals,
						Kind:  chain.KindRequest,
						Key:   nativeschema.PropertyKeyActorPublicKey,
						Value: hex.EncodeToString(forPutGet.Bytes()),
					},
				},
			},
		},
	}

	deniesGetPutOpForOthers := chain.Chain{
		Rules: []chain.Rule{
			{
				Status: chain.AccessDenied,
				Actions: chain.Actions{
					Names: []string{nativeschema.MethodAddTreeNode,
						nativeschema.MethodAddTreeNodeByPath,
						nativeschema.MethodRemoveTreeNode,
						nativeschema.MethodMoveTreeNode,
						nativeschema.MethodGetTreeNodeByPath,
						nativeschema.MethodGetSubtree,
						nativeschema.MethodListTrees},
				},
				Resources: chain.Resources{
					Names: []string{nativeschema.ResourceFormatRootTrees},
				},
				Condition: []chain.Condition{
					{
						Op:    chain.CondStringEquals,
						Kind:  chain.KindRequest,
						Key:   nativeschema.PropertyKeyActorRole,
						Value: nativeschema.PropertyValueContainerRoleOthers,
					},
					{
						Op:    chain.CondStringNotEquals,
						Kind:  chain.KindRequest,
						Key:   nativeschema.PropertyKeyActorPublicKey,
						Value: hex.EncodeToString(forPutGet.Bytes()),
					},
					{
						Op:    chain.CondStringNotEquals,
						Kind:  chain.KindRequest,
						Key:   nativeschema.PropertyKeyActorPublicKey,
						Value: hex.EncodeToString(forGet.Bytes()),
					},
				},
			},
		},
	}

	b.SetAPEOverride(bearer.APEOverride{
		Target: ape.ChainTarget{
			TargetType: ape.TargetTypeContainer,
			Name:       cid.EncodeToString(),
		},
		Chains: []ape.Chain{
			{
				Raw: allowsGetOpForAnyRoleForBothKeys.Bytes(),
			},
			{
				Raw: allowsPutOpForAnyRoleForOneKeyOnly.Bytes(),
			},
			{
				Raw: deniesGetPutOpForOthers.Bytes(),
			},
		},
	})

	return b
}
